library('FNN')  # This package is required

EkNNclus <- function(x,D,K,y0,N=1,q=0.5,sym=FALSE,p=1,disp=TRUE,tr=FALSE,cred=TRUE){
  
# EkNNclus algorithm (version: 14 March 2016)
# Source: 
#  T. Denoeux, O. Kanjanatarakul and S. Sriboonchitta. 
#  EK-NNclus: a clustering procedure based on the evidential K-nearest neighbor rule. 
#  Knowledge-Based Systems, Vol. 88, pages 57–69, 2015.
  
# INPUTS
# x: nxp data matrix (n instances, p attributes)
# D: nxn dissimilarity matrix (use donly if x is not supplied)
# K: number of neighbors
# y0: initial partition (vector of length n, with values in {1,2,...})
# N: number of runs of the algorithm (the best solution is kept)
# q:  parameter in (0,1). Gamma is set to the inverse of the q quantile of distances from the K-NN 
#     (same notation as in the paper)  
# sym: if TRUE, we enforce symmetry by setting 
#     alpha_ij = alpha_ji = phi(d_ij^p) if j in N_K(i) and i in N_K(j), and alpha_ij = 0
#     By default, sym=FALSE
# p: exponent of distances, alpha_ij = phi(d_ij^p) 
# disp: if TRUE, intermediate results are displayed
# tr: if TRUE, a trace of the objective function is returned  
# cred= if TRUE, the masses are returned (otherwise, only the final partition is returned)

# OUTPUTS
# y: the final partition (vector of length n)
# m: the credal partition (matrix nx(c+1), where c is the number of clusters)
# crit: the final value of the cost function
# trace: trace of the algorithm
# cpu: computing time
# W: the weight matrix
  

  if(missing(x)) nn<-knn_dist(D,K) else nn<-get.knn(x,K)
  eknn<-construct_W_eknn(nn,q,sym,p)
  cl<-cluster_hopfield.kNN(eknn$W,eknn$index,y0,N,disp,tr,cred)
  return(list(y=cl$y,m=cl$m,crit=cl$crit,trace=cl$trace,cpu=cl$cpu,W=eknn$W))
}

#---------
# The basic algorithm
#--------
cluster_hopfield.kNN<- function(W,index,y0,N=1,disp=TRUE,tr=FALSE,cred=TRUE){
  ptm<-proc.time()
  n<-dim(W)[1]
  critmax<- -1
  Trace<- NULL
  for(ntrials in (1:N)){
    nb.change<-1
    k<-0
    y<-y0
    c<-max(y)
    I<-diag(c);
    crit<-0
    for(i in (1:n)) crit<-crit+0.5*sum(W[i,]*(y[i]==y[index[i,]]))
    if(tr==TRUE) Trace<-rbind(Trace, c(ntrials,k,crit,c))
    while(nb.change>0){
      k<-k+1
      ii<-sample.int(n);
      nb.change<-0
      for(i in (1:n)){
        S<-I[y[index[ii[i],]],]
        u<-W[ii[i],]%*%S
        kstar<-which.max(u)
        crit<-crit+u[kstar]-u[y[ii[i]]]
        if(y[ii[i]] != kstar){
          y[ii[i]]<-kstar
          nb.change<-nb.change+1
        }
        if(tr==TRUE) Trace<-rbind(Trace, c(ntrials,k,crit,c)) 
      }
      y<-simplify.labels(y)
      c<-max(y)
      I<-diag(c);
      if(disp==TRUE) print(c(ntrials,k,nb.change,c))
    }
    if(crit>critmax){
      critmax<-crit
      ybest<-y
    }
  }
  m<-NULL
  if(cred==TRUE){
    alpha<-1-exp(-W)
    m<-credal.partition(ybest,alpha,index)
  }
  return(list(y=ybest,m=m,crit=critmax,trace=Trace,cpu=proc.time()-ptm))
}


# Construction of W matrix
construct_W_eknn<-function(nn,q=0.5,sym=FALSE,p=1){ 
	K<-dim(nn$nn.index)[2]
	n<-dim(nn$nn.index)[1]
	g<-1/quantile(nn$nn.dist^p,q)
	alpha<-pmax(exp(-g*nn$nn.dist^p),1e-6)
	if(sym==TRUE){
		for(i in (1:n)){
			for(j in (1:K)){
				if(!is.element(i,nn$nn.index[nn$nn.index[i,j],])) alpha[i,j]<- 0
			}
		}
	} 	
	W=-log(1-alpha) 
	return(list(alpha=alpha,W=W,index=nn$nn.index,sym=sym))
}

# Search of K NN from D matrix
knn_dist<- function(D,K){
	n<-dim(D)[1]
	knn<-matrix(0,nrow=n,ncol=K)
	index<-knn
	for(i in (1:n)){
		ds<-sort(D[i,],index.return=TRUE)
		knn[i,]<-ds$x[2:(K+1)]
		index[i,]<-ds$ix[2:(K+1)]
	}
	return(list(nn.dist=knn,nn.index=index))	
}

	


# Simplification of labels within the Hopfield procedure
simplify.labels<- function(y){
	n<-length(y)
	s<-sort(y,index.return=TRUE)
	ys<-s$x
	y1<-vector(length=n)
	y1[s$ix[1]]<-1
	for(i in (2:n)){
		if(ys[i]==ys[i-1]) y1[s$ix[i]]<-y1[s$ix[i-1]] else y1[s$ix[i]]<-y1[s$ix[i-1]]+1
	}
	return(y1)
}

# Computation of credal partition with Hopfield procedure
credal.partition <- function(y,alpha,index){
	n<-length(y)
	c<-max(y)
	K<-dim(alpha)[2]
	m<-cbind(matrix(0,n,c),rep(1,n))
	for(i in (1:n)){
		for(j in (1:K)){
			m1<-rep(0,c+1)
			k<-y[index[i,j]]
			m1[k] <- alpha[i,j]
			m1[c+1] <- 1-alpha[i,j]
			m[i,1:c] <- m1[1:c]*(m[i,1:c]+m[i,c+1])+ m1[c+1]*m[i,1:c]
			m[i,c+1]<- m1[c+1]*m[i,c+1]
			m[i,]<-m[i,]/sum(m[i,])
		}
	}
	return(m)
}

